! *****************************************************************
!
! This file contains a Fortran 90 module for computing the numerical 
! solution
!
! Copyright (C) 2010 David McInerney, University of Chicago
!
! *****************************************************************
!
!    This program is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with this program (see file "COPYING").  If not, see
!    <http://www.gnu.org/licenses/>.
!
! *****************************************************************

module numerical
  
  use parameters, only : period, LL, h0, BB, no_warmups
  use grid, only : JJ, NN, zeta, u_old, u_new, HH, x_at_zeta, x_at_u, &
       delta_x, delta_t, t, KK, v_old, v_new, x_at_v, y_at_v, delta_y, &
       y_at_zeta, y_at_u, iteration
  
  implicit none

#ifdef HAVE_NETCDF
  include 'netcdf.inc'
#endif

  double precision rx, ry
  logical :: advection_on = .true.
  
contains

! ***************************************************************

  subroutine setup_numerical
    
    use output_module, only : open_files

    implicit none
    
    integer j,k

! setup some constants
    delta_t = period/dble(NN)
    delta_x = LL/dble(JJ)
    delta_y = BB/dble(KK)
    rx = delta_t/delta_x             
    ry = delta_t/delta_y            
    
    print*
    print*,"JJ = ",JJ,": delta_x =",delta_x
    print*,"KK = ",KK,": delta_y =",delta_y
    print*,"NN = ",NN,": delta_t =",delta_t
    print*

! here we setup values for x and y at the zeta, u and v positions 

    do j=0,JJ
       x_at_u(j) = delta_x*dble(j)
    end do
    
    do j=1,JJ
       x_at_zeta(j) = delta_x*(dble(j)-0.5d0)
       x_at_v(j) = delta_x*(dble(j)-0.5d0)
    end do
    
    do k=0,KK
       y_at_v(k) = delta_y*dble(k)
    end do
    
    do k=1,KK
       y_at_zeta(k) = delta_y*(dble(k)-0.5d0)
       y_at_u(k) = delta_y*(dble(k)-0.5d0)
    end do
    
    call open_files

  end subroutine setup_numerical

! ***************************************************************

  subroutine run_numerical

    use differences, only : print_differences, reset_differences, &
         numerical_differences
    use analytic, only: analytic_zeta_at_zeta, analytic_u_at_u, &
         analytic_v_at_v, analytic_zeta1_at_zeta, analytic_u1_at_u, &
         analytic_v1_at_v
    use output_module, only : write_zeta_u_v, ncid, output_per_period
    
    integer j,k,n,number,m
    integer :: diff_freq = 20
    integer :: p
    integer rcode

! initial conditions. we use a cold start.
    do j = 1,JJ
       do k = 1,KK
          zeta(j,k) = 0d0
       end do
    end do
    zeta(1:JJ,1:KK) = 0.0d0
    u_old = 0.0d0
    u_new = 0.0d0
    v_old = 0d0
    v_new = 0d0
    HH = h0 + zeta

    iteration = 0

    print*,"Beginning warmup period"

! run model for no_warmups*period

    do p=1,no_warmups/diff_freq
       do m=1,diff_freq-1
          do n=1,NN
             call one_iteration
          end do
       end do
       call reset_differences
       do n=1,NN
          call one_iteration
          if (mod(dble(n),dble(NN)/dble(output_per_period)) .eq. 0d0) then
             call numerical_differences
          end if 
       end do
       call print_differences
    end do

    print*,"Warmup complete"
    
! reset differences before calculating sums of differences
    call reset_differences

! run model for another 2 periods, calculating differences and
! writing variables to file
    call write_zeta_u_v
    do n=1,NN
       call one_iteration
! print variables at locations specified earlier and calculate differences
       if (mod(dble(n),dble(NN)/dble(output_per_period)) .eq. 0d0) then
	  call numerical_differences
          call write_zeta_u_v
       end if
    end do

    print*,"Finished calculating differences and outputting data"

#ifdef HAVE_NETCDF
    call ncclos(ncid, rcode)
    print*,"Closed netcdf file"
#endif

! compute and print final differences
    call print_differences

  end subroutine run_numerical

! ***************************************************************

  subroutine one_iteration

! ***************************************************************

    implicit none
    
    integer j, k, ierr

    do j=1,JJ
       do k=1,KK
          call one_zeta(j,k)
       end do
    end do

    HH = h0 + zeta

    do j=1,JJ-1
       do k=1,KK
          call one_u(j,k)
       end do
    end do

    do j=1,JJ
       do k=1,KK-1
          call one_v(j,k)
       end do
    end do

    u_old = u_new
    v_old = v_new

    iteration = iteration + 1

    t = dble(iteration)*delta_t

    do j=1,JJ
       do k=1,KK
          if (abs(u_new(j,k)).gt.1000d0) then
             print*,"Stopped due to unstable results"
             stop
          end if
       end do
    end do
    
  end subroutine one_iteration

! ***************************************************************
! fd formula for the contuity eqn.

  subroutine one_zeta(j,k)

    implicit none

    integer j,k

    zeta(j,k) = zeta(j,k) - rx*(u_old(j,k)-u_old(j-1,k)) &
         - ry*(v_old(j,k)-v_old(j,k-1))

  end subroutine one_zeta

! ***************************************************************
! fd formula for the x-momentum eqn

  subroutine one_u(j,k)
    
    use parameters, only : g, rho, Cl
    use wind, only : wind_stress_x
    
    implicit none
    
    integer j,k
    double precision gravity_term, usq_advection_term, &
         uv_advection_term, wind_term, friction_term
    
    gravity_term = -0.5d0*g/delta_x*(HH(j,k)+HH(j+1,k)) &
         *(zeta(j+1,k)-zeta(j,k))

    if (advection_on) then    

      usq_advection_term =  -0.25d0/delta_x &
           *((u_old(j,k)+u_old(j+1,k))**2/HH(j+1,k) &
           -(u_old(j-1,k)+u_old(j,k))**2/HH(j,k))
    
      uv_advection_term = -1d0/delta_y &
           *((u_old(j,k)+u_old(j,k+1))*(v_old(j,k)+v_old(j+1,k)) &
           /(HH(j,k)+HH(j+1,k)+HH(j,k+1)+HH(j+1,k+1)) &
           -(u_old(j,k-1)+u_old(j,k))*(v_old(j,k-1)+v_old(j+1,k-1)) &
           /(HH(j,k-1)+HH(j+1,k-1)+HH(j,k)+HH(j+1,k)))
    
    else

      usq_advection_term = 0d0

      uv_advection_term = 0d0

    end if
 
    friction_term = -4d0*Cl*u_old(j,k)/(HH(j,k)+HH(j+1,k))**2
    
    wind_term = wind_stress_x(x_at_u(j),y_at_u(k),t+0.5d0*delta_t)/rho
    
    u_new(j,k) = u_old(j,k) + delta_t*(gravity_term &
         + usq_advection_term + uv_advection_term + wind_term &
         + friction_term)
    
  end subroutine one_u

! ***************************************************************
! fd formula for the y-momentum eqn

  subroutine one_v(j,k)
    
    use parameters, only : g, rho, Cl
    use wind, only : wind_stress_y
    
    implicit none
    
    integer j,k
    double precision gravity_term, uv_advection_term, &
         vsq_advection_term, wind_term, friction_term 
    
    gravity_term = -0.5d0*g/delta_y*(HH(j,k)+HH(j,k+1)) &
         *(zeta(j,k+1)-zeta(j,k))

    if (advection_on) then

      uv_advection_term = -1d0/delta_x &
           *((v_old(j,k)+v_old(j+1,k))*(u_old(j,k)+u_old(j,k+1)) &
           /(HH(j,k)+HH(j,k+1)+HH(j+1,k)+HH(j+1,k+1)) &
           -(v_old(j-1,k)+v_old(j,k))*(u_old(j-1,k)+u_old(j-1,k+1)) &
           /(HH(j-1,k)+HH(j-1,k+1)+HH(j,k)+HH(j,k+1)))
    
      vsq_advection_term =  -0.25d0/delta_y &
           *((v_old(j,k)+v_old(j,k+1))**2/HH(j,k+1) &
           -(v_old(j,k-1)+v_old(j,k))**2/HH(j,k))
    
    else

      uv_advection_term = 0d0

      vsq_advection_term = 0d0

    end if

    friction_term = -4d0*Cl*v_old(j,k)/(HH(j,k)+HH(j,k+1))**2
    
    wind_term = wind_stress_y(x_at_v(j),y_at_v(k),t+0.5d0*delta_t)/rho
    
    v_new(j,k) = v_old(j,k) + delta_t*(gravity_term  &
         + uv_advection_term + vsq_advection_term + wind_term &
         + friction_term)
    
  end subroutine one_v

! ********************************************************************

end module numerical
